#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2025 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: CC0-1.0

import os
import sys

from importlib.metadata import version as get_version

from packaging.version import parse as parse_version

os.environ["SPHINX_AUTODOC_RELOAD_MODULES"] = "1"

project = "culsans"
author = "Ilya Egorov"
copyright = "2025 Ilya Egorov"

v = parse_version(get_version("culsans"))
version = v.base_version
release = v.public

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "sphinx_rtd_theme",
]

if sys.version_info >= (3, 11):
    extensions.append("sphinxcontrib.autodoc_inherit_overload")

autodoc_class_signature = "separated"
autodoc_inherit_docstrings = False
autodoc_preserve_defaults = True
autodoc_default_options = {
    "exclude-members": "__init_subclass__,__class_getitem__,__weakref__",
    "inherited-members": True,
    "member-order": "bysource",
    "show-inheritance": True,
    "special-members": True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

html_theme = "sphinx_rtd_theme"
html_theme_options = {}
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_context = {
    "display_github": True,
    "github_user": "x42005e1f",
    "github_repo": "culsans",
    "github_version": "main",
    "conf_py_path": "/docs/",
}


def enable_api_index(app, docname, source):
    if docname == "api":
        source[0] = source[0].replace(":no-index:", "")


def setup(app):
    app.connect("source-read", enable_api_index)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

from __future__ import annotations

from typing import TypeVar

from ._protocols import AsyncQueue, MixedQueue, SyncQueue

_T = TypeVar("_T")


class SyncQueueProxy(SyncQueue[_T]):
    """
    A proxy that implements the :class:`SyncQueue` protocol by wrapping a mixed
    queue.
    """

    __slots__ = ("wrapped",)

    wrapped: MixedQueue[_T]

    def __init__(self, wrapped: MixedQueue[_T]) -> None:
        self.wrapped = wrapped

    def __repr__(self) -> str:
        cls = self.__class__
        cls_repr = f"{cls.__module__}.{cls.__qualname__}"

        return f"{cls_repr}({self.wrapped!r})"

    def peekable(self) -> bool:
        return self.wrapped.peekable()

    def qsize(self) -> int:
        return self.wrapped.qsize()

    def empty(self) -> bool:
        return self.wrapped.empty()

    def full(self) -> bool:
        return self.wrapped.full()

    def put(
        self,
        item: _T,
        block: bool = True,
        timeout: float | None = None,
    ) -> None:
        self.wrapped.sync_put(item, block, timeout)

    def put_nowait(self, item: _T) -> None:
        self.wrapped.put_nowait(item)

    def get(self, block: bool = True, timeout: float | None = None) -> _T:
        return self.wrapped.sync_get(block, timeout)

    def get_nowait(self) -> _T:
        return self.wrapped.get_nowait()

    def peek(self, block: bool = True, timeout: float | None = None) -> _T:
        return self.wrapped.sync_peek(block, timeout)

    def peek_nowait(self) -> _T:
        return self.wrapped.peek_nowait()

    def join(self) -> None:
        self.wrapped.sync_join()

    def task_done(self) -> None:
        self.wrapped.task_done()

    def shutdown(self, immediate: bool = False) -> None:
        self.wrapped.shutdown(immediate)

    def clear(self) -> None:
        self.wrapped.clear()

    @property
    def unfinished_tasks(self) -> int:
        return self.wrapped.unfinished_tasks

    @property
    def is_shutdown(self) -> bool:
        return self.wrapped.is_shutdown

    @property
    def closed(self) -> bool:
        return self.wrapped.closed

    @property
    def maxsize(self) -> int:
        return self.wrapped.maxsize

    @maxsize.setter
    def maxsize(self, value: int) -> None:
        self.wrapped.maxsize = value


class AsyncQueueProxy(AsyncQueue[_T]):
    """
    A proxy that implements the :class:`AsyncQueue` protocol by wrapping a
    mixed queue.
    """

    __slots__ = ("wrapped",)

    wrapped: MixedQueue[_T]

    def __init__(self, wrapped: MixedQueue[_T]) -> None:
        self.wrapped = wrapped

    def __repr__(self) -> str:
        cls = self.__class__
        cls_repr = f"{cls.__module__}.{cls.__qualname__}"

        return f"{cls_repr}({self.wrapped!r})"

    def peekable(self) -> bool:
        return self.wrapped.peekable()

    def qsize(self) -> int:
        return self.wrapped.qsize()

    def empty(self) -> bool:
        return self.wrapped.empty()

    def full(self) -> bool:
        return self.wrapped.full()

    async def put(self, item: _T) -> None:
        await self.wrapped.async_put(item)

    def put_nowait(self, item: _T) -> None:
        self.wrapped.put_nowait(item)

    async def get(self) -> _T:
        return await self.wrapped.async_get()

    def get_nowait(self) -> _T:
        return self.wrapped.get_nowait()

    async def peek(self) -> _T:
        return await self.wrapped.async_peek()

    def peek_nowait(self) -> _T:
        return self.wrapped.peek_nowait()

    async def join(self) -> None:
        await self.wrapped.async_join()

    def task_done(self) -> None:
        self.wrapped.task_done()

    def shutdown(self, immediate: bool = False) -> None:
        self.wrapped.shutdown(immediate)

    def clear(self) -> None:
        self.wrapped.clear()

    @property
    def unfinished_tasks(self) -> int:
        return self.wrapped.unfinished_tasks

    @property
    def is_shutdown(self) -> bool:
        return self.wrapped.is_shutdown

    @property
    def closed(self) -> bool:
        return self.wrapped.closed

    @property
    def maxsize(self) -> int:
        return self.wrapped.maxsize

    @maxsize.setter
    def maxsize(self, value: int) -> None:
        self.wrapped.maxsize = value



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

from __future__ import annotations

import sys

from collections import deque
from heapq import heappop, heappush
from math import inf, isnan
from typing import Any, Protocol, TypeVar, Union

from aiologic import Condition
from aiologic.lowlevel import async_checkpoint, green_checkpoint

from ._exceptions import (
    QueueEmpty,
    QueueFull,
    QueueShutDown,
    UnsupportedOperation,
)
from ._protocols import MixedQueue
from ._proxies import AsyncQueueProxy, SyncQueueProxy

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

try:
    from aiologic.lowlevel import ThreadLock, create_thread_lock
except ImportError:  # aiologic<0.15.0
    from aiologic.lowlevel._thread import (  # type: ignore[assignment]
        LockType as ThreadLock,
        allocate_lock as create_thread_lock,
    )

try:
    from aiologic.lowlevel import green_clock
except ImportError:  # aiologic<0.15.0
    from time import monotonic as green_clock

try:
    from aiologic.lowlevel import (
        async_checkpoint_enabled,
        green_checkpoint_enabled,
    )
except ImportError:  # aiologic<0.15.0

    def async_checkpoint_enabled() -> bool:
        return True

    def green_checkpoint_enabled() -> bool:
        return True


_T = TypeVar("_T")
_T_contra = TypeVar("_T_contra", contravariant=True)


class _SupportsBool(Protocol):
    __slots__ = ()

    def __bool__(self, /) -> bool: ...


class _SupportsLT(Protocol[_T_contra]):
    __slots__ = ()

    def __lt__(self, other: _T_contra, /) -> _SupportsBool: ...


class _SupportsGT(Protocol[_T_contra]):
    __slots__ = ()

    def __gt__(self, other: _T_contra, /) -> _SupportsBool: ...


_RichComparableT = TypeVar(
    "_RichComparableT",
    bound=Union[_SupportsLT[Any], _SupportsGT[Any]],
)


class Queue(MixedQueue[_T]):
    """
    A mixed sync-async queue that is:

    * :abbr:`MPMC (multi-producer, multi-consumer)`
    * :abbr:`FIFO (first-in, first-out)`

    Compliant with the Janus API version 2.0.0.
    """

    __slots__ = (
        "__data",
        "__weakref__",
        "_is_shutdown",
        "_maxsize",
        "_unfinished_tasks",
        "all_tasks_done",
        "mutex",
        "not_empty",
        "not_full",
    )

    __data: deque[_T]

    _maxsize: int
    _unfinished_tasks: int
    _is_shutdown: bool

    mutex: ThreadLock

    not_full: Condition[ThreadLock]
    not_empty: Condition[ThreadLock]
    all_tasks_done: Condition[ThreadLock]

    def __init__(self, maxsize: int = 0) -> None:
        """
        Create a queue object with the given maximum size.

        if *maxsize* is <= 0, the queue size if infinite. If it is an integer
        greater that 0, then the put methods block when the queue reaches
        *maxsize* until an item is removed by the get methods.
        """

        self._maxsize = maxsize
        self._unfinished_tasks = 0
        self._is_shutdown = False

        self.mutex = mutex = create_thread_lock()

        self.not_full = Condition(mutex)  # putters
        self.not_empty = Condition(mutex)  # getters
        self.all_tasks_done = Condition(mutex)  # joiners

        self._init(maxsize)  # data

    def peekable(self) -> bool:
        with self.mutex:
            return self._peekable()

    def qsize(self) -> int:
        with self.mutex:
            return self._qsize()

    def empty(self) -> bool:
        with self.mutex:
            return not self._qsize()

    def full(self) -> bool:
        with self.mutex:
            return 0 < self._maxsize <= self._qsize()

    def sync_put(
        self,
        item: _T,
        block: bool = True,
        timeout: float | None = None,
    ) -> None:
        rescheduled = False
        endtime = None

        with self.mutex:
            while True:
                self._check_closing()

                if 0 < self._maxsize:
                    if not block:
                        if 0 < self._maxsize <= self._qsize():
                            raise QueueFull
                    elif timeout is None:
                        while 0 < self._maxsize <= self._qsize():
                            self.not_full.wait()

                            self._check_closing()

                            rescheduled = True
                    else:
                        if isinstance(timeout, int):
                            try:
                                timeout = float(timeout)
                            except OverflowError:
                                timeout = (-1 if timeout < 0 else 1) * inf

                        if isnan(timeout):
                            msg = "'timeout' must be a number (non-NaN)"
                            raise ValueError(msg)

                        if timeout < 0:
                            msg = "'timeout' must be a non-negative number"
                            raise ValueError(msg)

                        if endtime is None:
                            endtime = green_clock() + timeout

                        while 0 < self._maxsize <= self._qsize():
                            remaining = endtime - green_clock()

                            if remaining <= 0:
                                raise QueueFull

                            self.not_full.wait(remaining)

                            self._check_closing()

                            rescheduled = True

                if not rescheduled and green_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        green_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            self._put(item)
            self._unfinished_tasks += 1

            self.not_empty.notify()

    async def async_put(self, item: _T) -> None:
        rescheduled = False

        with self.mutex:
            while True:
                self._check_closing()

                while 0 < self._maxsize <= self._qsize():
                    await self.not_full

                    self._check_closing()

                    rescheduled = True

                if not rescheduled and async_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        await async_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            self._put(item)
            self._unfinished_tasks += 1

            self.not_empty.notify()

    def put_nowait(self, item: _T) -> None:
        with self.mutex:
            self._check_closing()

            if 0 < self._maxsize <= self._qsize():
                raise QueueFull

            self._put(item)
            self._unfinished_tasks += 1

            self.not_empty.notify()

    def sync_get(self, block: bool = True, timeout: float | None = None) -> _T:
        rescheduled = False
        endtime = None

        with self.mutex:
            while True:
                if not block:
                    if not self._qsize():
                        self._check_closing()

                        raise QueueEmpty
                elif timeout is None:
                    while not self._qsize():
                        self._check_closing()

                        self.not_empty.wait()

                        rescheduled = True
                else:
                    if isinstance(timeout, int):
                        try:
                            timeout = float(timeout)
                        except OverflowError:
                            timeout = (-1 if timeout < 0 else 1) * inf

                    if isnan(timeout):
                        msg = "'timeout' must be a number (non-NaN)"
                        raise ValueError(msg)

                    if timeout < 0:
                        msg = "'timeout' must be a non-negative number"
                        raise ValueError(msg)

                    if endtime is None:
                        endtime = green_clock() + timeout

                    while not self._qsize():
                        self._check_closing()

                        remaining = endtime - green_clock()

                        if remaining <= 0:
                            raise QueueEmpty

                        self.not_empty.wait(remaining)

                        rescheduled = True

                if not rescheduled and green_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        green_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            item = self._get()

            if 0 >= self._maxsize or self._maxsize > self._qsize():
                self.not_full.notify()

        return item

    async def async_get(self) -> _T:
        rescheduled = False

        with self.mutex:
            while True:
                while not self._qsize():
                    self._check_closing()

                    await self.not_empty

                    rescheduled = True

                if not rescheduled and async_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        await async_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            item = self._get()

            if 0 >= self._maxsize or self._maxsize > self._qsize():
                self.not_full.notify()

        return item

    def get_nowait(self) -> _T:
        with self.mutex:
            if not self._qsize():
                self._check_closing()

                raise QueueEmpty

            item = self._get()

            if 0 >= self._maxsize or self._maxsize > self._qsize():
                self.not_full.notify()

        return item

    def sync_peek(
        self,
        block: bool = True,
        timeout: float | None = None,
    ) -> _T:
        rescheduled = False
        notified = False
        endtime = None

        with self.mutex:
            self._check_peekable()

            while True:
                if not block:
                    if not self._qsize():
                        self._check_closing()

                        raise QueueEmpty
                elif timeout is None:
                    while not self._qsize():
                        self._check_closing()

                        notified = self.not_empty.wait()

                        rescheduled = True
                else:
                    if isinstance(timeout, int):
                        try:
                            timeout = float(timeout)
                        except OverflowError:
                            timeout = (-1 if timeout < 0 else 1) * inf

                    if isnan(timeout):
                        msg = "'timeout' must be a number (non-NaN)"
                        raise ValueError(msg)

                    if timeout < 0:
                        msg = "'timeout' must be a non-negative number"
                        raise ValueError(msg)

                    if endtime is None:
                        endtime = green_clock() + timeout

                    while not self._qsize():
                        self._check_closing()

                        remaining = endtime - green_clock()

                        if remaining <= 0:
                            raise QueueEmpty

                        notified = self.not_empty.wait(remaining)

                        rescheduled = True

                if not rescheduled and green_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        green_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            try:
                item = self._peek()
            finally:
                if notified:
                    self.not_empty.notify()

        return item

    async def async_peek(self) -> _T:
        rescheduled = False
        notified = False

        with self.mutex:
            self._check_peekable()

            while True:
                while not self._qsize():
                    self._check_closing()

                    notified = await self.not_empty

                    rescheduled = True

                if not rescheduled and async_checkpoint_enabled():
                    self.mutex.release()

                    try:
                        await async_checkpoint()
                    finally:
                        self.mutex.acquire()

                    rescheduled = True
                else:
                    break

            try:
                item = self._peek()
            finally:
                if notified:
                    self.not_empty.notify()

        return item

    def peek_nowait(self) -> _T:
        with self.mutex:
            self._check_peekable()

            if not self._qsize():
                self._check_closing()

                raise QueueEmpty

            item = self._peek()

        return item

    def sync_join(self) -> None:
        rescheduled = False

        with self.mutex:
            while self._unfinished_tasks:
                self.all_tasks_done.wait()

                rescheduled = True

        if not rescheduled:
            green_checkpoint()

    async def async_join(self) -> None:
        rescheduled = False

        with self.mutex:
            while self._unfinished_tasks:
                await self.all_tasks_done

                rescheduled = True

        if not rescheduled:
            await async_checkpoint()

    def task_done(self) -> None:
        with self.mutex:
            unfinished = self._unfinished_tasks - 1

            if unfinished <= 0:
                if unfinished < 0:
                    msg = "task_done() called too many times"
                    raise ValueError(msg)

                self.all_tasks_done.notify_all()

            self._unfinished_tasks = unfinished

    def shutdown(self, immediate: bool = False) -> None:
        with self.mutex:
            self._is_shutdown = True

            if immediate:
                while self._qsize():
                    self._get()

                    if self._unfinished_tasks > 0:
                        self._unfinished_tasks -= 1

                self.all_tasks_done.notify_all()

            self.not_empty.notify_all()
            self.not_full.notify_all()

    def close(self) -> None:
        """
        Close the queue.

        This method is provided for compatibility with the Janus queues. Use
        :meth:`queue.shutdown(immediate=True) <BaseQueue.shutdown>` as a direct
        substitute.
        """

        self.shutdown(immediate=True)

    async def wait_closed(self) -> None:
        """
        Wait for finishing all pending activities.

        This method is provided for compatibility with the Janus queues. It
        actually does nothing.

        Raises:
          RuntimeError:
            if called for non-closed queue.
        """

        if not self._is_shutdown:
            msg = "Waiting for non-closed queue"
            raise RuntimeError(msg)

        await async_checkpoint()

    async def aclose(self) -> None:
        """
        Shutdown the queue and wait for actual shutting down.

        This method is provided for compatibility with the Janus queues. Use
        :meth:`queue.shutdown(immediate=True) <BaseQueue.shutdown>` as a direct
        substitute.
        """

        self.close()
        await self.wait_closed()

    def clear(self) -> None:
        with self.mutex:
            size = self._qsize()
            unfinished = max(0, self._unfinished_tasks - size)

            self._clear()

            if not unfinished:
                self.all_tasks_done.notify_all()

            self._unfinished_tasks = unfinished
            self.not_full.notify(size)

    def _check_peekable(self) -> None:
        if not self._peekable():
            msg = "peeking not supported"
            raise UnsupportedOperation(msg)

    def _check_closing(self) -> None:
        if self._is_shutdown:
            raise QueueShutDown

    # Override these methods to implement other queue organizations
    # (e.g. stack or priority queue).
    # These will only be called with appropriate locks held

    def _init(self, maxsize: int) -> None:
        self.__data = deque()

    def _qsize(self) -> int:
        return len(self.__data)

    def _put(self, item: _T) -> None:
        self.__data.append(item)

    def _get(self) -> _T:
        return self.__data.popleft()

    def _peek(self) -> _T:
        return self.__data[0]

    def _peekable(self) -> bool:
        return True

    def _clear(self) -> None:
        self.__data.clear()

    @property
    def sync_q(self) -> SyncQueueProxy[_T]:
        return SyncQueueProxy(self)

    @property
    def async_q(self) -> AsyncQueueProxy[_T]:
        return AsyncQueueProxy(self)

    @property
    def putting(self) -> int:
        """
        The current number of threads/tasks waiting to put.

        It represents the length of the wait queue and thus changes
        immediately.
        """

        return self.not_full.waiting

    @property
    def getting(self) -> int:
        """
        The current number of threads/tasks waiting to get/peek.

        It represents the length of the wait queue and thus changes
        immediately.
        """

        return self.not_empty.waiting

    @property
    def waiting(self) -> int:
        """
        The current number of threads/tasks waiting to access.

        It is roughly equivalent to the sum of the :attr:`putting` and
        :attr:`getting` properties, but is more reliable than the sum in a
        multithreaded environment.
        """

        with self.mutex:
            # We use the underlying lock to ensure that no thread can increment
            # the counters during normal queue operation (since exclusive
            # access is required to enter the wait queue).
            return self.not_full.waiting + self.not_empty.waiting

    @property
    def unfinished_tasks(self) -> int:
        return self._unfinished_tasks

    @property
    def is_shutdown(self) -> bool:
        return self._is_shutdown

    @property
    def closed(self) -> bool:
        return self._is_shutdown

    @property
    def maxsize(self) -> int:
        return self._maxsize

    @maxsize.setter
    def maxsize(self, value: int) -> None:
        with self.mutex:
            maxsize = self._maxsize

            if value <= 0:
                self.not_full.notify_all()
            elif value > maxsize:
                self.not_full.notify(value - maxsize)

            self._maxsize = value


class LifoQueue(Queue[_T]):
    """
    A variant of :class:`Queue` that retrieves most recently added entries
    first (:abbr:`LIFO (last-in, first-out)`).
    """

    __slots__ = ("__data",)  # noqa: PLW0244

    __data: list[_T]

    @override
    def _init(self, maxsize: int) -> None:
        self.__data = []

    @override
    def _qsize(self) -> int:
        return len(self.__data)

    @override
    def _put(self, item: _T) -> None:
        self.__data.append(item)

    @override
    def _get(self) -> _T:
        return self.__data.pop()

    @override
    def _peek(self) -> _T:
        return self.__data[-1]

    @override
    def _peekable(self) -> bool:
        return True

    @override
    def _clear(self) -> None:
        self.__data.clear()


class PriorityQueue(Queue[_RichComparableT]):
    """
    A variant of :class:`Queue` that retrieves entries in priority order
    (lowest first).
    """

    __slots__ = ("__data",)  # noqa: PLW0244

    __data: list[_RichComparableT]

    @override
    def _init(self, maxsize: int) -> None:
        self.__data = []

    @override
    def _qsize(self) -> int:
        return len(self.__data)

    @override
    def _put(self, item: _RichComparableT) -> None:
        heappush(self.__data, item)

    @override
    def _get(self) -> _RichComparableT:
        return heappop(self.__data)

    @override
    def _peek(self) -> _RichComparableT:
        return self.__data[0]

    @override
    def _peekable(self) -> bool:
        return True

    @override
    def _clear(self) -> None:
        self.__data.clear()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: 0BSD

import asyncio

import pytest

import culsans


class TestMixedQueue:
    factory = culsans.Queue

    @pytest.mark.asyncio
    async def test_sync_put_async_get(self):
        queue = self.factory()
        loop = asyncio.get_running_loop()

        def sync_run(sync_q):
            assert sync_q.empty()

            for i in range(5):
                sync_q.put(i)

        async def async_run(async_q):
            for i in range(5):
                assert await async_q.get() == i

            assert async_q.empty()

        for _ in range(3):
            await asyncio.gather(
                loop.run_in_executor(None, sync_run, queue.sync_q),
                async_run(queue.async_q),
            )

    @pytest.mark.asyncio
    async def test_async_put_sync_get(self):
        queue = self.factory()
        loop = asyncio.get_running_loop()

        def sync_run(sync_q):
            for i in range(5):
                assert sync_q.get() == i

            assert sync_q.empty()

        async def async_run(async_q):
            assert async_q.empty()

            for i in range(5):
                await async_q.put(i)

        for _ in range(3):
            await asyncio.gather(
                loop.run_in_executor(None, sync_run, queue.sync_q),
                async_run(queue.async_q),
            )

    @pytest.mark.asyncio
    async def test_sync_join_async_done(self):
        queue = self.factory()
        loop = asyncio.get_running_loop()

        def sync_run(sync_q):
            assert sync_q.empty()

            for i in range(5):
                sync_q.put(i)

            sync_q.join()

        async def async_run(async_q):
            for i in range(5):
                assert await async_q.get() == i

                async_q.task_done()

            assert async_q.empty()

        for _ in range(3):
            await asyncio.gather(
                loop.run_in_executor(None, sync_run, queue.sync_q),
                async_run(queue.async_q),
            )

    @pytest.mark.asyncio
    async def test_async_join_sync_done(self):
        queue = self.factory()
        loop = asyncio.get_running_loop()

        def sync_run(sync_q):
            for i in range(5):
                assert sync_q.get() == i

                sync_q.task_done()

            assert sync_q.empty()

        async def async_run(async_q):
            assert async_q.empty()

            for i in range(5):
                await async_q.put(i)

            await async_q.join()

        for _ in range(3):
            await asyncio.gather(
                loop.run_in_executor(None, sync_run, queue.sync_q),
                async_run(queue.async_q),
            )

    @pytest.mark.asyncio
    async def test_growing_and_shrinking(self):
        queue = self.factory(1)
        loop = asyncio.get_running_loop()

        queue.async_q.put_nowait(0)
        assert queue.async_q.qsize() == 1

        task = loop.run_in_executor(None, queue.sync_q.put, 1)

        while not queue.putting:
            await asyncio.sleep(1e-3)

        queue.async_q.maxsize = 2  # growing

        await task
        assert queue.async_q.qsize() == 2

        task = loop.run_in_executor(None, queue.sync_q.put, 2)

        while not queue.putting:
            await asyncio.sleep(1e-3)

        queue.async_q.maxsize = 1  # shrinking

        assert queue.putting == 1
        assert queue.async_q.qsize() == 2

        queue.async_q.get_nowait()

        assert queue.putting == 1
        assert queue.async_q.qsize() == 1

        queue.async_q.maxsize = 0  # now the queue size is infinite

        await task
        assert queue.async_q.qsize() == 2

    @pytest.mark.asyncio
    async def test_peek(self):
        queue = self.factory()

        assert queue.sync_q.peekable()
        assert queue.async_q.peekable()

        with pytest.raises(culsans.SyncQueueEmpty):
            queue.sync_q.peek(block=False)
        with pytest.raises(culsans.SyncQueueEmpty):
            queue.sync_q.peek_nowait()
        with pytest.raises(culsans.AsyncQueueEmpty):
            queue.async_q.peek_nowait()

        queue.async_q.put_nowait(42)

        assert queue.sync_q.peek() == 42
        assert queue.sync_q.peek_nowait() == 42
        assert await queue.async_q.peek() == 42
        assert queue.async_q.peek_nowait() == 42

    @pytest.mark.asyncio
    async def test_clear(self):
        queue = self.factory()
        loop = asyncio.get_running_loop()

        for i in range(3):
            queue.async_q.put_nowait(i)

        assert queue.unfinished_tasks == 3
        assert queue.async_q.qsize() == 3

        task = loop.run_in_executor(None, queue.sync_q.join)

        queue.async_q.clear()

        assert queue.unfinished_tasks == 0
        assert queue.async_q.qsize() == 0

        await task

        for i in range(3):
            queue.async_q.put_nowait(i)

        assert queue.unfinished_tasks == 3
        assert queue.async_q.qsize() == 3

        task = asyncio.create_task(queue.async_q.join())

        queue.sync_q.clear()

        assert queue.unfinished_tasks == 0
        assert queue.async_q.qsize() == 0

        await task

    @pytest.mark.asyncio
    async def test_closed(self):
        queue = self.factory()

        assert not queue.closed
        assert not queue.async_q.closed
        assert not queue.sync_q.closed

        queue.close()

        assert queue.closed
        assert queue.async_q.closed
        assert queue.sync_q.closed

    @pytest.mark.asyncio
    async def test_double_closing(self):
        queue = self.factory()

        queue.close()
        queue.close()

        await queue.wait_closed()

    @pytest.mark.asyncio
    async def test_wait_without_closing(self):
        queue = self.factory()

        with pytest.raises(RuntimeError):
            await queue.wait_closed()

        queue.close()
        await queue.wait_closed()

    @pytest.mark.asyncio
    async def test_modifying_forbidden_after_closing(self):
        queue = self.factory()
        queue.close()

        with pytest.raises(culsans.SyncQueueShutDown):
            queue.sync_q.put(5)
        with pytest.raises(culsans.SyncQueueShutDown):
            queue.sync_q.put_nowait(5)

        with pytest.raises(culsans.SyncQueueShutDown):
            queue.sync_q.get()
        with pytest.raises(culsans.SyncQueueShutDown):
            queue.sync_q.get_nowait()

        with pytest.raises(culsans.AsyncQueueShutDown):
            await queue.async_q.put(5)
        with pytest.raises(culsans.AsyncQueueShutDown):
            queue.async_q.put_nowait(5)

        with pytest.raises(culsans.AsyncQueueShutDown):
            await queue.async_q.get()
        with pytest.raises(culsans.AsyncQueueShutDown):
            queue.async_q.get_nowait()

        await queue.wait_closed()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

"""Thread-safe async-aware queue for Python"""

from ._exceptions import (
    AsyncQueueEmpty as AsyncQueueEmpty,
    AsyncQueueFull as AsyncQueueFull,
    AsyncQueueShutDown as AsyncQueueShutDown,
    QueueEmpty as QueueEmpty,
    QueueFull as QueueFull,
    QueueShutDown as QueueShutDown,
    SyncQueueEmpty as SyncQueueEmpty,
    SyncQueueFull as SyncQueueFull,
    SyncQueueShutDown as SyncQueueShutDown,
    UnsupportedOperation as UnsupportedOperation,
)
from ._protocols import (
    AsyncQueue as AsyncQueue,
    BaseQueue as BaseQueue,
    MixedQueue as MixedQueue,
    SyncQueue as SyncQueue,
)
from ._proxies import (
    AsyncQueueProxy as AsyncQueueProxy,
    SyncQueueProxy as SyncQueueProxy,
)
from ._queues import (
    LifoQueue as LifoQueue,
    PriorityQueue as PriorityQueue,
    Queue as Queue,
)

# update .__module__ attributes for shorter representation
__import__(f"{__name__}._utils", fromlist=["_export"])._export(globals())



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

from __future__ import annotations

from typing import Protocol, TypeVar

from ._utils import _copydoc as copydoc

_T = TypeVar("_T")


class BaseQueue(Protocol[_T]):
    """
    A base queue protocol that includes all except blocking methods.

    Can be useful when you need to make sure that none of the methods will
    block (except for the underlying lock).
    """

    __slots__ = ()

    def peekable(self) -> bool:
        """
        Return :data:`True` if the queue is peekable, :data:`False` otherwise.
        """
        ...

    def qsize(self) -> int:
        """
        Return the number of items in the queue.

        This method is provided for compatibility with the standard queues.
        Note, ``queue.qsize() > 0`` does not guarantee that subsequent get/peek
        calls will not block. Similarly, ``queue.qsize() < queue.maxsize`` does
        not guarantee that subsequent put calls will not block.
        """
        ...

    def empty(self) -> bool:
        """
        Return :data:`True` if the queue is empty, :data:`False` otherwise.

        This method is provided for compatibility with the standard queues. Use
        ``queue.qsize() == 0`` as a direct substitute, but be aware that either
        approach risks a race condition where the queue can grow before the
        result of :meth:`empty` or :meth:`qsize` can be used.

        To create code that needs to wait for all queued tasks to be completed,
        the preferred technique is to use the join methods.
        """
        ...

    def full(self) -> bool:
        """
        Return :data:`True` if the queue is full, :data:`False` otherwise.

        This method is provided for compatibility with the standard queues. Use
        ``queue.qsize() >= queue.maxsize`` as a direct substitute, but be aware
        that either approach risks a race condition where the queue can shrink
        before the result of :meth:`full` or :meth:`qsize` can be used.
        """
        ...

    def put_nowait(self, item: _T) -> None:
        """
        Put *item* into the queue without blocking.

        Only put (enqueue) the item if a free slot is immediately available.

        Raises:
          QueueFull:
            if the queue is full.
          QueueShutDown:
            if the queue has been shut down.
        """
        ...

    def get_nowait(self) -> _T:
        """
        Remove and return an item from the queue without blocking.

        Only get (dequeue) an item if one is immediately available.

        Raises:
          QueueEmpty:
            if the queue is empty.
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
        """
        ...

    def peek_nowait(self) -> _T:
        """
        Return an item from the queue without blocking.

        Only peek (front) an item if one is immediately available.

        Raises:
          QueueEmpty:
            if the queue is empty.
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
          UnsupportedOperation:
            if the queue is not peekable.
        """
        ...

    def task_done(self) -> None:
        """
        Indicate that a formerly enqueued task is complete.

        Used by queue consumers.  For each get call used to fetch a task,
        a subsequent call to :meth:`task_done` tells the queue that the
        processing on the task is complete.

        If a join call is currently blocking, it will resume when all items
        have been processed (meaning that a :meth:`task_done` call was received
        for every item that had been put into the queue).

        Raises:
          ValueError:
            if called more times than there were items placed in the queue.
        """
        ...

    def shutdown(self, immediate: bool = False) -> None:
        """
        Put the queue into a shutdown mode.

        The queue can no longer grow. Future calls to the put methods raise
        :exc:`QueueShutDown`. Currently blocked callers of the put methods will
        be unblocked and will raise :exc:`QueueShutDown` in the formerly
        blocked thread/task.

        Once the queue is empty, the get/peek methods will also raise
        :exc:`QueueShutDown`.

        Args:
          immediate:
            If set to :data:`True`, the queue is drained to be completely
            empty, and :meth:`task_done` is called for each item removed from
            the queue (but joiners are unblocked regardless of the number of
            unfinished tasks).
        """
        ...

    def clear(self) -> None:
        """
        Clear all items from the queue atomically.

        Also calls :meth:`task_done` for each removed item.
        """
        ...

    @property
    def unfinished_tasks(self) -> int:
        """
        The current number of tasks remaining to be processed.

        See the :meth:`task_done` method.
        """
        ...

    @property
    def is_shutdown(self) -> bool:
        """
        A boolean that is :data:`True` if the queue has been shut down,
        :data:`False` otherwise.

        See the :meth:`shutdown` method.
        """
        ...

    @property
    def closed(self) -> bool:
        """
        A boolean that is :data:`True` if the queue has been closed,
        :data:`False` otherwise.

        This property is provided for compatibility with the Janus queues. Use
        :attr:`queue.is_shutdown <is_shutdown>` as a direct substitute.
        """
        ...

    @property
    def maxsize(self) -> int:
        """
        The maximum number of items which the queue can hold.

        It can be changed dynamically by setting the attribute.
        """
        ...

    @maxsize.setter
    def maxsize(self, value: int) -> None: ...


class MixedQueue(BaseQueue[_T], Protocol[_T]):
    """
    A mixed queue protocol that includes both types of blocking methods via
    prefixes.

    Provides specialized proxies via properties.
    """

    __slots__ = ()

    def sync_put(
        self,
        item: _T,
        block: bool = True,
        timeout: float | None = None,
    ) -> None:
        """
        Put *item* into the queue.

        Args:
          block:
            Unless set to :data:`False`, the method will block if necessary
            until a free slot is available. Otherwise, ``timeout=0`` is
            implied.
          timeout:
            If set to a non-negative number, the method will block at most
            *timeout* seconds and raise the :exc:`QueueFull` exception if no
            free slot was available within that time.

        Raises:
          QueueFull:
            if the queue is full and the timeout has expired.
          QueueShutDown:
            if the queue has been shut down.
          ValueError:
            if *timeout* is negative or :data:`NaN <math.nan>`.
        """
        ...

    async def async_put(self, item: _T) -> None:
        """
        Put *item* into the queue.

        If the queue is full, wait until a free slot is available.

        Raises:
          QueueShutDown:
            if the queue has been shut down.
        """
        ...

    def sync_get(
        self,
        block: bool = True,
        timeout: float | None = None,
    ) -> _T:
        """
        Remove and return an item from the queue.

        Args:
          block:
            Unless set to :data:`False`, the method will block if necessary
            until an item is available. Otherwise, ``timeout=0`` is implied.
          timeout:
            If set to a non-negative number, the method will block at most
            *timeout* seconds and raise the :exc:`QueueEmpty` exception if no
            item was available within that time.

        Raises:
          QueueEmpty:
            if the queue is empty and the timeout has expired.
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
          ValueError:
            if *timeout* is negative or :data:`NaN <math.nan>`.
        """
        ...

    async def async_get(self) -> _T:
        """
        Remove and return an item from the queue.

        If the queue is empty, wait until an item is available.

        Raises:
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
        """
        ...

    def sync_peek(
        self,
        block: bool = True,
        timeout: float | None = None,
    ) -> _T:
        """
        Return an item from the queue without removing it.

        Args:
          block:
            Unless set to :data:`False`, the method will block if necessary
            until an item is available. Otherwise, ``timeout=0`` is implied.
          timeout:
            If set to a non-negative number, the method will block at most
            *timeout* seconds and raise the :exc:`QueueEmpty` exception if no
            item was available within that time.

        Raises:
          QueueEmpty:
            if the queue is empty and the timeout has expired.
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
          UnsupportedOperation:
            if the queue is not peekable.
          ValueError:
            if *timeout* is negative or :data:`NaN <math.nan>`.
        """
        ...

    async def async_peek(self) -> _T:
        """
        Return an item from the queue without removing it.

        If the queue is empty, wait until an item is available.

        Raises:
          QueueShutDown:
            if the queue has been shut down and is empty, or if the queue has
            been shut down immediately.
          UnsupportedOperation:
            if the queue is not peekable.
        """
        ...

    def sync_join(self) -> None:
        """
        Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls
        :meth:`~BaseQueue.task_done` to indicate that the item was retrieved
        and all work on it is complete.

        When the count of unfinished tasks drops to zero, the caller unblocks.
        """
        ...

    async def async_join(self) -> None:
        """
        Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls
        :meth:`~BaseQueue.task_done` to indicate that the item was retrieved
        and all work on it is complete.

        When the count of unfinished tasks drops to zero, the caller unblocks.
        """
        ...

    @property
    def sync_q(self) -> SyncQueue[_T]:
        """
        An interface compatible with the standard queues from the :mod:`queue`
        module.
        """
        ...

    @property
    def async_q(self) -> AsyncQueue[_T]:
        """
        An interface compatible with the standard queues from the
        :mod:`asyncio` module.
        """
        ...


class SyncQueue(BaseQueue[_T], Protocol[_T]):
    """
    A synchronous queue protocol that covers the standard queues' interface
    from the :mod:`queue` module.

    Compliant with the Python API version 3.13.
    """

    __slots__ = ()

    @copydoc(MixedQueue.sync_put)
    def put(
        self,
        item: _T,
        block: bool = True,
        timeout: float | None = None,
    ) -> None: ...
    @copydoc(MixedQueue.sync_get)
    def get(self, block: bool = True, timeout: float | None = None) -> _T: ...
    @copydoc(MixedQueue.sync_peek)
    def peek(self, block: bool = True, timeout: float | None = None) -> _T: ...
    @copydoc(MixedQueue.sync_join)
    def join(self) -> None: ...


class AsyncQueue(BaseQueue[_T], Protocol[_T]):
    """
    An asynchronous queue protocol that covers the standard queues' interface
    from the :mod:`asyncio` module.

    Compliant with the Python API version 3.13.
    """

    __slots__ = ()

    @copydoc(MixedQueue.async_put)
    async def put(self, item: _T) -> None: ...
    @copydoc(MixedQueue.async_get)
    async def get(self) -> _T: ...
    @copydoc(MixedQueue.async_peek)
    async def peek(self) -> _T: ...
    @copydoc(MixedQueue.async_join)
    async def join(self) -> None: ...



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: 0BSD

import pytest

import culsans


class _TestQueueBase:
    def test_init(self):
        assert self.factory().maxsize == 0
        assert self.factory(1).maxsize == 1
        assert self.factory(maxsize=-1).maxsize == -1

    def test_getters(self):
        queue = self.factory()

        assert queue.sync_q.wrapped is queue
        assert queue.async_q.wrapped is queue

        assert queue.unfinished_tasks == 0
        assert queue.sync_q.unfinished_tasks == 0
        assert queue.async_q.unfinished_tasks == 0

        assert not queue.is_shutdown
        assert not queue.sync_q.is_shutdown
        assert not queue.async_q.is_shutdown

        assert queue.maxsize == 0
        assert queue.sync_q.maxsize == 0
        assert queue.async_q.maxsize == 0

        with pytest.raises(AttributeError):
            queue.nonexistent_attribute  # noqa: B018
        with pytest.raises(AttributeError):
            queue.sync_q.nonexistent_attribute  # noqa: B018
        with pytest.raises(AttributeError):
            queue.async_q.nonexistent_attribute  # noqa: B018

    def test_setters(self):
        queue = self.factory()

        with pytest.raises(AttributeError):
            queue.sync_q = queue.sync_q
        with pytest.raises(AttributeError):
            queue.async_q = queue.async_q

        with pytest.raises(AttributeError):
            queue.unfinished_tasks = queue.unfinished_tasks
        with pytest.raises(AttributeError):
            queue.sync_q.unfinished_tasks = queue.sync_q.unfinished_tasks
        with pytest.raises(AttributeError):
            queue.async_q.unfinished_tasks = queue.async_q.unfinished_tasks

        with pytest.raises(AttributeError):
            queue.is_shutdown = queue.is_shutdown
        with pytest.raises(AttributeError):
            queue.sync_q.is_shutdown = queue.sync_q.is_shutdown
        with pytest.raises(AttributeError):
            queue.async_q.is_shutdown = queue.async_q.is_shutdown

        queue.maxsize = 1
        assert queue.maxsize == 1
        queue.maxsize = -1
        assert queue.maxsize == -1

        queue.sync_q.maxsize = 1
        assert queue.maxsize == 1
        queue.sync_q.maxsize = -1
        assert queue.maxsize == -1

        queue.async_q.maxsize = 1
        assert queue.maxsize == 1
        queue.async_q.maxsize = -1
        assert queue.maxsize == -1

        with pytest.raises(AttributeError):
            queue.nonexistent_attribute = 42
        with pytest.raises(AttributeError):
            queue.sync_q.nonexistent_attribute = 42
        with pytest.raises(AttributeError):
            queue.async_q.nonexistent_attribute = 42

    def test_deleters(self):
        queue = self.factory()

        with pytest.raises(AttributeError):
            del queue.sync_q
        with pytest.raises(AttributeError):
            del queue.async_q

        with pytest.raises(AttributeError):
            del queue.unfinished_tasks
        with pytest.raises(AttributeError):
            del queue.sync_q.unfinished_tasks
        with pytest.raises(AttributeError):
            del queue.async_q.unfinished_tasks

        with pytest.raises(AttributeError):
            del queue.is_shutdown
        with pytest.raises(AttributeError):
            del queue.sync_q.is_shutdown
        with pytest.raises(AttributeError):
            del queue.async_q.is_shutdown

        with pytest.raises(AttributeError):
            del queue.maxsize
        with pytest.raises(AttributeError):
            del queue.sync_q.maxsize
        with pytest.raises(AttributeError):
            del queue.async_q.maxsize

        with pytest.raises(AttributeError):
            del queue.nonexistent_attribute
        with pytest.raises(AttributeError):
            del queue.sync_q.nonexistent_attribute
        with pytest.raises(AttributeError):
            del queue.async_q.nonexistent_attribute

    def test_unfinished(self):
        queue = self.factory()
        assert queue.unfinished_tasks == 0

        queue.sync_q.put(1)
        assert queue.unfinished_tasks == 1

        queue.sync_q.get()
        assert queue.unfinished_tasks == 1

        queue.sync_q.task_done()
        assert queue.unfinished_tasks == 0

    def test_shutdown(self):
        queue = self.factory()

        queue.sync_q.put(1)
        queue.sync_q.task_done()
        queue.sync_q.put(2)

        queue.shutdown()

        assert queue._qsize() == 2
        assert queue.unfinished_tasks == 1
        assert queue.is_shutdown

        queue.shutdown()

    def test_immediate_shutdown(self):
        queue = self.factory()

        queue.sync_q.put(1)
        queue.sync_q.task_done()
        queue.sync_q.put(2)

        queue.shutdown(immediate=True)

        assert queue._qsize() == 0
        assert queue.unfinished_tasks == 0
        assert queue.is_shutdown

        queue.shutdown(immediate=True)


class TestQueue(_TestQueueBase):
    factory = culsans.Queue


class TestLifoQueue(_TestQueueBase):
    factory = culsans.LifoQueue


class TestPriorityQueue(_TestQueueBase):
    factory = culsans.PriorityQueue



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

import sys

from asyncio import QueueEmpty as AsyncEmpty, QueueFull as AsyncFull
from queue import Empty as SyncEmpty, Full as SyncFull

if sys.version_info >= (3, 13):
    from asyncio import QueueShutDown as AsyncShutDown
    from queue import ShutDown as SyncShutDown

    AsyncQueueShutDown = AsyncShutDown
    SyncQueueShutDown = SyncShutDown

    class QueueShutDown(SyncQueueShutDown, AsyncQueueShutDown):
        """
        Raised when put/get/peek with shut-down queue.
        """

else:

    class QueueShutDown(Exception):
        """
        Raised when put/get/peek with shut-down queue.
        """

    SyncQueueShutDown = QueueShutDown
    AsyncQueueShutDown = QueueShutDown

AsyncQueueEmpty = AsyncEmpty
SyncQueueEmpty = SyncEmpty

AsyncQueueFull = AsyncFull
SyncQueueFull = SyncFull


class QueueFull(SyncQueueFull, AsyncQueueFull):
    """
    Raised when non-blocking put with full queue.
    """


class QueueEmpty(SyncQueueEmpty, AsyncQueueEmpty):
    """
    Raised when non-blocking get/peek with empty queue.
    """


class UnsupportedOperation(ValueError):
    """
    Raised when peek with non-peekable queue.
    """



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2025 Ilya Egorov <0x42005e1f@gmail.com>
# SPDX-License-Identifier: ISC

from __future__ import annotations

import sys

from typing import TYPE_CHECKING, Any, TypeVar

if sys.version_info >= (3, 9):
    from collections.abc import Callable
else:
    from typing import Callable

_CallableT = TypeVar("_CallableT", bound=Callable[..., Any])


def _copydoc(
    wrapped: Callable[..., Any],
) -> Callable[[_CallableT], _CallableT]:
    def decorator(func: _CallableT) -> _CallableT:
        func.__doc__ = wrapped.__doc__

        return func

    return decorator


def _export(namespace: dict[str, Any]) -> None:
    if TYPE_CHECKING:
        # sphinx.ext.autodoc does not support hacks below. In particular,
        # 'bysource' ordering will not work, nor will some cross-references. So
        # we skip them on type checking (implied by
        # SPHINX_AUTODOC_RELOAD_MODULES=1).
        return

    module_name: str = namespace["__name__"]

    for value in namespace.values():
        if getattr(value, "__module__", "").startswith(f"{module_name}."):
            value.__module__ = module_name
